 
(* ocamlopt -unsafe -o eric Types.ml Lex.ml Eric.ml *)

open Lex;; 
 
module Loop(ER:Types.ER_DiGraph)
=
struct
	let rec print_concept = function
		| ER.Concept(t) ->
			print_string t 
		| ER.ConceptN(t,n) ->
			print_string t; print_char ' '; print_string n
		| ER.ConceptS(t,s) ->
			print_string t; print_char ' '; 
			print_char '"'; print_string s; print_char '"'
		| ER.ConceptI(t,i) -> 
			print_string t; print_char ' '; print_int i
		| ER.ConceptD m | ER.ConceptV m ->
			print_concept m.ER.concept; print_char '*'; print_string m.ER.marker
		| ER.ConceptU m ->
			()
	let print_concept = function
		| ER.ConceptU m -> print_string m.ER.marker
		| c -> print_char '['; print_concept c; print_char ']'
	let rec print_pattern (p:ER.pattern) =
		match p with
		| [] -> print_char '.'; print_newline () 
		| (ca,r,cb)::tl ->
			print_char '('; print_concept ca; 
			print_char ' '; print_string r; print_char ' ';
			print_concept cb; print_char ')';
			print_pattern tl
	let print_pattern p () =
		print_pattern p
	let main =
		let module Dict = Types.Dictionary in
		object
			val mutable kb = ER.make ()
			val mutable dict = Dict.empty
			val mutable select = false
		initializer
			print_endline "Entity-Relationship Interactive Calculator"; 
			while true do
				print_string  "# "; flush stdout;
				let lex = new Lex.make (Lex.Channel stdin) in
				let demand_marker c =
					if not (lex#granted_mono '*') then c
					else if select then
						let marker = lex#demand_name in
						let d = ER.ConceptD {ER.concept=c;marker} in
						dict <- Dict.add marker d dict; d
					else
						(dict <- Dict.add lex#demand_name c dict; c)
				in
				let demand_concept () =
					match () with 
					| _ when lex#granted_mono '[' ->
						if select && lex#granted_mono '*' then
							let marker = lex#demand_name in
							lex#demand_mono ']';
							let v = ER.ConceptV {ER.concept=ER.empty;marker} in
							dict <- Dict.add marker v dict; v
						else
							let t = lex#demand_name in
							if not (lex#promised_mono ']') then
							ignore(lex#granted_mono ':');
							let c = match () with
							| _ when lex#promised_name ->
								ER.ConceptN(t,lex#demand_name)
							| _ when lex#granted_mono '#' ->
								ER.ConceptI(t,lex#demand_int)
							| _ when lex#granted_mono '+' ->
								ER.ConceptI(t,lex#demand_int)
							| _ when lex#promised_int ->
								ER.ConceptI(t,lex#demand_int)
							| _ when lex#granted_mono '-' ->
								ER.ConceptI(t,-lex#demand_int)
							| _ when lex#promised_mono '"' ->
								ER.ConceptS(t,lex#demand_string)
							| _ -> ER.Concept(t)						
							in let d = demand_marker c in lex#demand_mono ']'; d
					| _ ->
						let c = Dict.find lex#demand_name dict in
						if not select then c else match c with
						| ER.ConceptD m | ER.ConceptV m -> ER.ConceptU m
						| _ -> c 
				in
				let rec demand_join () =
					lex#demand_mono '(';
					let ca = demand_concept ()
					and r  = lex#demand_name
					and cb = demand_concept ()
					in
					lex#demand_mono ')';
					ER.insert ca r cb kb;
					if lex#promised_mono '(' then demand_join ()
				in
				let rec demand_select () =
					lex#demand_mono '(';
					let ca = demand_concept ()
					and r  = lex#demand_name
					and cb = demand_concept ()
					in
					lex#demand_mono ')';
					(ca,r,cb)::
					(if lex#promised_mono '(' then demand_select () else [])
				in
				let assert_mono c =
					if not (lex#promised_mono c) then lex#demand_mono c
				in
				try
					dict <- Dict.empty; select <- false; 
					match () with
					| _ when lex#granted_word "join" -> 
						demand_join (); assert_mono '.'
					| _ when lex#granted_word "select" ->
						select <- true; 
						let pat = demand_select () in
						assert_mono '.';
						ER.success := print_pattern pat;
						ER.select pat kb
					| _ when lex#granted_word "load" -> 
						assert_mono '"';
						begin
							try
								let str = lex#demand_string in
								assert_mono '.';
								let ch = open_in_bin str in
								kb <- (Marshal.from_channel ch : ER.t);
								close_in ch
							with Sys_error _ -> failwith "can not read file"
						end
					| _ when lex#granted_word "save" -> 
						assert_mono '"';
						begin
							try
								let str = lex#demand_string in
								assert_mono '.';
								let ch = open_out_bin str in
								Marshal.to_channel ch kb ER.marshal_flags;
								close_out ch
							with Sys_error _ -> failwith "can not write file"
						end
					| _ when lex#granted_word "quit" -> 
						assert_mono '.'; exit 0
					| _ -> failwith "command expected"
				with
				| Lex.Demand_denied (p,d) -> Lex.print_position p; Lex.print_denied d
				| Failure msg -> print_endline msg
				| End_of_file -> print_endline "unexpected end of text"
				| Not_found   -> print_endline "unkown identifier"
			done 
		end 
end

module Main = Loop(Types.ER_Diagram(Types.Diagram))

